#include <sys/unistd.h> 
#include <stdio.h> 
#include <stdlib.h> 

// mouvements du robot
#define GAUCHE 1 
#define DROITE 2 
#define HAUT 3 
#define BAS 4 


#include "node.hpp"
#include "pg.hpp"

//parcours
int table[64] = { 
0,0,0,0,0,1,1,1, 
0,0,0,0,1,2,2,2, 
0,0,0,1,1,0,0,0, 
0,0,1,1,1,1,1,0, 
0,0,0,1,1,0,0,0, 
0,0,0,0,0,0,0,0, 
0,0,0,0,0,0,0,0, 
0,0,0,0,0,0,0,0}; 

int verifie_table[64]; 

//position de depart du robot
int depart = 28;

// nombre max de generations 
int mg =  500;

// definition de la classe noeud 
class NodeAspir : public Node
{
 public:


  NodeAspir(){};
  ~NodeAspir(){};

  void aff_nodes()
  {
    switch(element) 
      { 
      case GAUCHE: printf("(GAUCHE)"); break; 
      case DROITE: printf("(DROITE)"); break; 
      case HAUT: printf("(HAUT)"); break; 
      case BAS: printf("(BAS)"); break; 
      default: printf("(ERREUR)"); break; 
      } 
    if (fils != NULL) fils->aff_nodes(); 
  }

NodeAspir * clone_nodes()
  // clone la structure de noeuds
{ 
  NodeAspir * res; 
  res = new NodeAspir();
  res->element = element;
  if (fils != NULL) {
  res->fils = fils->clone_nodes();
  }
  else res->fils = NULL;
  return res; 
}

  
  
};


// definition de la classe PgAspir: prog genetique du robot
class PgAspir : public Pg
{
public:
  PgAspir(){}
  ~PgAspir(){}

private:

 Node * creer_arbre(int prof)
  // creer un genome aleatoire
  {
    NodeAspir *p; 
    p = new NodeAspir();
    p->element = (rand() % 4)+1; 
    if (prof <=0) 
	p->fils = NULL; 
else 
	p->fils = creer_arbre(prof-1); 
return p; 
  }

  int verifie_arbre(Node * p,int pos) 
  { 
    if (pos<0) return 0; 
    if (pos>63) return 0; 
    if (table[pos] == 2) return 0; 
    if (p == NULL) return 1; 
    if ((pos % 8 == 0) && (p->element == GAUCHE)) return 0; 
    if ((pos % 7 == 0) && (p->element == DROITE)) return 0; 
    switch(p->element) 
      { 
      case GAUCHE: return(verifie_arbre(p->fils,pos-1)); 
      case DROITE: return(verifie_arbre(p->fils,pos+1)); 
      case HAUT: return(verifie_arbre(p->fils,pos-8)); 
      case BAS: return(verifie_arbre(p->fils,pos+8)); 
      } 
    // ne doit pas se terminer ainsi 
    return 0; 
  } 

  float nb_cases_faites(Node * p,int pos) 
    // calcul le nombre de cases effectuees
  {  
    int r; 
    int newpos; 
 
    r = verifie_table[pos]; 
    if (r) verifie_table[pos]--; 
 
    if (p == NULL) return r; 
 
    switch(p->element) 
      { 
      case GAUCHE: newpos = pos-1; break; 
      case DROITE: newpos = pos+1; break; 
      case HAUT: newpos = pos - 8; break; 
      case BAS: newpos = pos + 8; break; 
      } 
 
    return (r+nb_cases_faites(p->fils,newpos)); 
  } 
   
  void fitness(pindividu pi)
    // calcul le fitness de l'individu
  {
    float res; 
    int i; 
 
    for (i=0;i<64;i++) { verifie_table[i] = table[i]; } 
 
    if (verifie_arbre(pi->pg,depart)) { 
      pi->fitness = Node::calc_nb_noeuds(pi->pg)+1000/(nb_cases_faites(pi->pg,depart)); 
    } 
    else pi->fitness = MaxFitness(); 

  }

  
	 
};

void open_datafile(char * file)
  // charge le fichier de donnes (map/point de depart/generations max)
  {
    FILE *f;
    int i;
 
    f=fopen(file,"r,");
    if (f == NULL) { printf("File %s not found !!\n",file); return; }
    fscanf(f,"maxgen=%d\n",&mg);
    fscanf(f,"depart=%d\n",&depart);
    fscanf(f,"(%d,",&table[0]);
    for (i=1;i<63;i++) fscanf(f,"%d,",&table[i]);
    fscanf(f,"%d),",&table[63]);
    fclose(f);
}

void print_map(void) 
{
  // affiche la map
  int i; 
  for(i=0;i<64;i++) 
    { 
      if (i%8 == 0) printf("\n"); 
      if (table[i] == 2) printf("- "); else printf("%d ",table[i]); 
    } 
  printf("\n"); 
  sleep(5); 
} 
 
 
void print_result(Node  * p,int pos) 
  // affiche en mouvement le parcours du robot pour un programme genetique
{ 
  int i,newpos; 
 
  if (table[pos]) table[pos]--; 
  for(i=0;i<64;i++) 
    { 
      if (i%8 == 0) printf("\n"); 
      if (i == pos) printf("X "); else if (table[i] == 2) printf("- "); else printf("%d ",table[i]); 
    }   
 printf("\n"); 
 
 if (p == NULL) return; 
 sleep(1); 
 switch(p->element) 
 { 
 case GAUCHE: newpos = pos-1; break; 
 case DROITE: newpos = pos+1; break; 
 case HAUT: newpos = pos - 8; break; 
 case BAS: newpos = pos + 8; break; 
 } 
 print_result(p->fils,newpos); 
} 


/* PROGRAMME PRINCIPAL */


int main()
{
  PgAspir * prog;

  open_datafile("aspir.dat");
  prog = new PgAspir();
  prog->SetMaxGen(mg);
  prog->cycle_pg();
  print_result(prog->GetSuperIndividu()->pg,depart);
  return 0;
}
